from dataclasses import dataclass, fields
from typing import Dict, List

__all__ = ["DuplaApiKeys"]


@dataclass
class DuplaApiKeys:
    """Container for DUPLA API keys."""

    CVR: str = "VirksomhedCVRNummer"
    SE: str = "VirksomhedSENummer"
    CPR: str = "PersonCPRNummer"

    TEKNISK_REGISTRERING_FRA: str = "TekniskRegistreringDatoFra"
    TEKNISK_REGISTRERING_TIL: str = "TekniskRegistreringDatoTil"

    AFREGNING_START: str = "AfregningPeriodeForholdPeriodeStartDato"
    AFREGNING_SLUT: str = "AfregningPeriodeForholdPeriodeSlutDato"

    UDSTILLING_FRA: str = "UdstillingRegistreringFra"
    UDSTILLING_TIL: str = "UdstillingRegistreringTil"

    ANGIVELSE_GYLDIG_FRA: str = "AngivelseGyldigDatoFra"
    ANGIVELSE_GYLDIG_TIL: str = "AngivelseGyldigDatoTil"

    REGISTRERING_FORHOLD_FRA: str = "RegistreringForholdGyldigDatoFra"
    REGISTRERING_FORHOLD_TIL: str = "RegistreringForholdGyldigDatoFra"

    STATUS_GYLDIG_FRA: str = "VirksomhedStatusTypeGyldigFra"
    STATUS_GYLDIG_TIL: str = "VirksomhedStatusTypeGyldigTil"

    SELVANGIVESE_INDKOMST_AAR: str = "SelskabSelvangivelseIndkomstÅr"

    SAG_TYPE_KODE: str = "SagTypeKode"
    PLIGT_KODE: str = "PligtKode"
    STATUS_TYPE_KODE: str = "VirksomhedStatusTypeKode"

    @classmethod
    def get_all_field_names(cls) -> List[str]:
        """Get all of the available attributes."""
        return [f.name for f in fields(cls)]

    @classmethod
    def get_all_fields(cls) -> Dict[str, str]:
        """Get a dictionary mapping the class attributes to the attribute values."""
        return {f: getattr(cls, f) for f in cls.get_all_field_names()}



from datetime import date, datetime

import pytest

from dupla import validation

# Constant dates, ensures they do not change during test
TODAY = date.today()
TODAY_DT = datetime.now()


@pytest.mark.parametrize(
    "date_inp,is_valid",
    [
        ("2012-12-12", True),
        ("2012-13-12", False),
        ("03-05-2020", False),
        ("01-01-20", False),
    ],
)
def test_validate_iso(date_inp, is_valid):
    assert validation.validate_iso_date(date_inp) == is_valid


@pytest.mark.parametrize(
    "date_inp,expected",
    [
        (datetime.strptime("12-03-2020", "%d-%m-%Y"), "2020-03-12"),
        (TODAY_DT, TODAY_DT.strftime("%Y-%m-%d")),
        (TODAY, TODAY.strftime("%Y-%m-%d")),
    ],
)
def test_convert_and_validate_iso(date_inp, expected):
    output = validation.convert_and_validate_iso_date(date_inp)
    assert isinstance(output, str)
    assert output == expected


@pytest.mark.parametrize("date_inp", ["03-05-2020", "2017-13-01"])
def test_invalid_iso(date_inp):
    with pytest.raises(ValueError):
        validation.convert_and_validate_iso_date(date_inp)



import abc
from datetime import datetime
from typing import Any, Dict, Optional
from urllib.parse import urljoin

from pydantic import BaseModel, ConfigDict, Field, field_serializer

from .api_keys import DuplaApiKeys
from .timestamp import as_utc

ALIAS_MAPPING: Dict[str, str] = {
    "default_endpoint": "default_endpoint",
    "se": DuplaApiKeys.SE,
    "cvr": DuplaApiKeys.CVR,
    "cpr": DuplaApiKeys.CPR,
    "registrering_fra": DuplaApiKeys.TEKNISK_REGISTRERING_FRA,
    "registrering_til": DuplaApiKeys.TEKNISK_REGISTRERING_TIL,
    "afregning_start": DuplaApiKeys.AFREGNING_START,
    "afregning_slut": DuplaApiKeys.AFREGNING_SLUT,
    "selvangivelse_aar": DuplaApiKeys.SELVANGIVESE_INDKOMST_AAR,
    "angivelse_fra": DuplaApiKeys.ANGIVELSE_GYLDIG_FRA,
    "angivelse_til": DuplaApiKeys.ANGIVELSE_GYLDIG_TIL,
    "udstilling_fra": DuplaApiKeys.UDSTILLING_FRA,
    "udstilling_til": DuplaApiKeys.UDSTILLING_TIL,
    "regisrering_forhold_fra": DuplaApiKeys.REGISTRERING_FORHOLD_FRA,
    "regisrering_forhold_til": DuplaApiKeys.REGISTRERING_FORHOLD_TIL,
    "status_gyldig_fra": DuplaApiKeys.STATUS_GYLDIG_FRA,
    "status_gyldig_til": DuplaApiKeys.STATUS_GYLDIG_TIL,
    # Codes
    "pligt_kode": DuplaApiKeys.PLIGT_KODE,
    "status_type_kode": DuplaApiKeys.STATUS_TYPE_KODE,
    "sag_type_kode": DuplaApiKeys.SAG_TYPE_KODE,
}


def _get_alias(name: str) -> str:
    """Get the mapping between the Pydantic field name and the Dupla key name."""
    return ALIAS_MAPPING.get(name, name)


class BasePayload(BaseModel, abc.ABC):
    """Base Payload Pydantic model"""

    model_config = ConfigDict(alias_generator=_get_alias, populate_by_name=True, extra="forbid")

    @property
    @abc.abstractmethod
    def default_endpoint(self) -> str:
        """The default endpoint for the payload"""
        # Solution from: https://github.com/pydantic/pydantic/discussions/2410#discussioncomment-408613

    def get_payload(self) -> Dict[str, Any]:
        """Get the payload in a json-able format.
        Excludes None values."""
        return self.model_dump(
            mode="json",
            exclude=["default_endpoint"],
            by_alias=True,
            exclude_none=True,
        )

    @classmethod
    def endpoint_from_base_url(cls, base_url: str) -> str:
        if not cls.default_endpoint:
            raise ValueError(f"The API {cls.__name__} has not set the default endpoint.")
        return urljoin(base_url, cls.default_endpoint)


class UdstillingMixin(BaseModel, abc.ABC):
    udstilling_fra: Optional[datetime] = Field(
        default=None,
    )
    udstilling_til: Optional[datetime] = Field(
        default=None,
    )

    @field_serializer("udstilling_fra", "udstilling_til")
    def serialize_udstilling(self, dt: Optional[datetime]) -> Optional[str]:
        if dt is None:
            return dt
        return as_utc(dt).strftime("%Y-%m-%dT%H:%M:%SZ")



import logging
from typing import Any, Dict, List, Optional

import backoff
import requests

from .base import DuplaApiBase
from .exceptions import DuplaApiException, DuplaResponseException
from .payload import BasePayload

logger = logging.getLogger(__file__)

__all__ = ["DuplaAccess"]

RESPONSE_T = Dict[str, Any]


class DuplaAccess(DuplaApiBase):
    """
    Class for accessing the Dataudveklspingsplatformen API (Dupla).
    """

    def __init__(
        self,
        transaction_id: str,
        agreement_id: str,
        pkcs12_filename: str,
        pkcs12_password: str,
        billetautomat_url: str,
        base_url: str = r"https://api.skat.dk",
        jwt_token_expiration_overlap: int = 5,
        max_tries: int = 8,
    ):
        """Instantiates new DUPLA API endpoint client.
        Args:
            base_url (str): The HTTP(S) endpoint of the API.
            transaction_id (str): An ID used to correlate requests across the API.
                Should be constant for IKP-DA.
            agreement_id (str): An ID/token supplied by the API provider.
            pkcs12_filename (str): Path to PKCS12 certificate file.
            pkcs12_password (str): Password for PKCS12 certificate file.
            billetautomat_url (str): Endpoint to the authentication service for requesting
                JWT tokens.
            jwt_token_expiration_overlap (int): The overlap time for token expiration time
                (in seconds) to avoid situations where token is almost expired during the check
                and will be rejected in a next request. Defaults to 5 seconds.
            max_tries (int): Maximum number of times a failed request is re-attempted in
                ``get_data``. Defaults to 8.
        """

        self.base_url = base_url
        self.max_tries = max_tries
        super().__init__(
            transaction_id,
            agreement_id,
            pkcs12_filename,
            pkcs12_password,
            billetautomat_url,
            jwt_token_expiration_overlap,
        )

    def get_endpoint(self, payload: BasePayload) -> str:
        """Retrieve the endpoint URL."""
        return payload.endpoint_from_base_url(self.base_url)

    def get_data(
        self,
        payload: BasePayload,
        endpoint: Optional[str] = None,
    ) -> List[RESPONSE_T]:
        """Request the server for data.

        Args:
            payload (BasePayload): The Pydantic payload model.
            endpoint (Optional[str], optional): An optional endpoint URL override.
                If not provided, it defaults to the url join of the base URL and
                the payload default URL. Defaults to None.
        Returns:
            List[Dict[str, Any]]: A JSON list representing data as returned by the API.
        """
        if endpoint is None:
            endpoint = self.get_endpoint(payload)
        payload_serialized = payload.get_payload()
        return self._run_payload(payload_serialized, endpoint)

    def _run_payload(self, payload: Dict[str, Any], endpoint: str) -> List[RESPONSE_T]:
        """Execute a given payload. No conversion is done on the payload."""

        # Construct the getter with a backoff, and a modified number of max tries
        @backoff.on_exception(
            backoff.expo,
            (requests.exceptions.RequestException, DuplaApiException),
            max_tries=self.max_tries,
        )
        def _getter():
            response = self.get(endpoint, params=payload)
            response.raise_for_status()

            try:
                response_json: Dict[str, Any] = response.json()

                # Perform simple type check to fail fast if the server has returned
                # something unknown.
                data = response_json["data"]
                if not isinstance(data, list):
                    logger.exception(
                        "Received an invalid response from DUPLA, which was not a list: %s", data
                    )
                    raise DuplaResponseException(
                        "Invalid response from DUPLA. The data key did not contain a list.",
                        response=response,
                    )
                return data
            except DuplaResponseException as e:
                # Let the inner exception through
                raise e
            except Exception as e:
                logger.exception("Error occurred while processing response: %s", response.content)
                raise DuplaResponseException(
                    "An error occurred while parsing the DUPLA response.",
                    response=response,
                ) from e

        return _getter()



from datetime import date
from typing import ClassVar, List, Optional

from pydantic import Field

from .abstract_payload import BasePayload, UdstillingMixin
from .custom_types import CPR_T, CVR_T, SE_T

ENDP_T = ClassVar[str]  # Endpoint type


class KtrPayload(BasePayload):
    """An API client for Dataudstillingsplatformens (DUPLA) Kontrolregistreringer API."""

    default_endpoint: ENDP_T = "Kontrolregistreringer/Virksomhed"

    cvr: Optional[CVR_T] = Field(default=None)
    se: Optional[SE_T] = Field(default=None)
    registrering_fra: Optional[date] = Field(default=None)
    registrering_til: Optional[date] = Field(default=None)


class KtrObsPayload(BasePayload):
    """An API client for Dataudstillingsplatformens (DUPLA) Kontrolobservationer API."""

    default_endpoint: ENDP_T = "Kontrolobservationer"

    cpr: CPR_T = Field()
    registrering_fra: Optional[date] = Field(default=None)
    registrering_til: Optional[date] = Field(default=None)


class LigPayload(BasePayload):
    """An API client for Dataudstillingsplatformens (DUPLA) Ligningssager API."""

    default_endpoint: ENDP_T = "Ligningssager"

    cvr: Optional[CVR_T] = Field(default=None)
    se: Optional[SE_T] = Field(default=None)
    cpr: Optional[CPR_T] = Field(default=None)

    sag_type_kode: Optional[str] = Field(default=None)

    registrering_fra: Optional[date] = Field(default=None)
    registrering_til: Optional[date] = Field(default=None)


class MomsPayload(UdstillingMixin, BasePayload):
    """An API client for Dataudstillingsplatformens (DUPLA) Momsangivelser API."""

    default_endpoint: ENDP_T = "Momsangivelse"
    se: SE_T = Field()
    afregning_start: date = Field()
    afregning_slut: date = Field()


class LonsumPayload(BasePayload):
    """An API client for Dataudstillingsplatformens (DUPLA) Lønsumsangivelser API."""

    default_endpoint: ENDP_T = "Lønsumsangivelser"
    se: SE_T = Field()

    registrering_fra: Optional[date] = Field(default=None)
    registrering_til: Optional[date] = Field(default=None)

    angivelse_fra: Optional[date] = Field(default=None)
    angivelse_til: Optional[date] = Field(default=None)


class SelskabSambeskatningPayload(BasePayload):
    """An API client for Dataudstillingsplatformens (DUPLA) Selskabsambeskatningskreds API."""

    default_endpoint: ENDP_T = "Selskabsskatteoplysninger/Selskabsambeskatningskreds"
    cvr: Optional[CVR_T] = Field(default=None)
    se: Optional[SE_T] = Field(default=None)

    selvangivelse_aar: Optional[str] = Field(default=None)
    registrering_fra: Optional[date] = Field(default=None)
    registrering_til: Optional[date] = Field(default=None)


class SelskabSelvangivelsePayload(BasePayload):
    """An API client for Dataudstillingsplatformens (DUPLA) Selskabselvangivelse API."""

    default_endpoint: ENDP_T = "Selskabsskatteoplysninger/Selskabselvangivelse"

    cvr: Optional[CVR_T] = Field(default=None)
    se: Optional[SE_T] = Field(default=None)

    selvangivelse_aar: Optional[str] = Field(default=None)
    registrering_fra: Optional[date] = Field(default=None)
    registrering_til: Optional[date] = Field(default=None)


class VirksomhedspligterPayload(BasePayload):
    """A payload for accessing DUPLA Virksomhedspligter."""

    default_endpoint: ENDP_T = "Virksomhedspligter"

    cvr: Optional[CVR_T] = Field(default=None)
    se: Optional[SE_T] = Field(default=None)

    pligt_kode: Optional[List[int]] = Field(default=None)

    regisrering_forhold_fra: Optional[date] = Field(default=None)
    regisrering_forhold_til: Optional[date] = Field(default=None)

    registrering_fra: Optional[date] = Field(default=None)
    registrering_til: Optional[date] = Field(default=None)


class VirksomhedsstatusPayload(BasePayload):
    """A payload for accessing DUPLA Virksomhedsstatus."""

    default_endpoint: ENDP_T = "Virksomhedsstatus"

    cvr: Optional[CVR_T] = Field(default=None)
    se: Optional[SE_T] = Field(default=None)

    status_type_kode: Optional[List[int]] = Field(default=None)

    status_gyldig_fra: Optional[date] = Field(default=None)
    status_gyldig_til: Optional[date] = Field(default=None)

    registrering_fra: Optional[date] = Field(default=None)
    registrering_til: Optional[date] = Field(default=None)



import uuid

import dupla
import dupla.version
from dupla.base import DuplaApiBase


def test_token_refreshed_if_expired(mocked_requests_very_short_expiration_time, mocker):
    mock_session_post, mock_session_request = mocked_requests_very_short_expiration_time
    api = DuplaApiBase(
        str(uuid.uuid4()),
        str(uuid.uuid4()),
        "pkcs12_filename",
        "pkcs12_password",
        "http://billetautomat.dk/url",
        5,
    )
    authentication_spy = mocker.spy(api, "_authenticate")

    api.get("http://some_api.dk/url")
    api.get("http://some_api.dk/url")

    assert authentication_spy.call_count == 2
    assert mock_session_post.call_count == 2
    assert mock_session_request.call_count == 2


def test_token_not_refreshed_if_not_expired(mocked_requests_long_expiration_time, mocker):
    mock_session_post, mock_session_request = mocked_requests_long_expiration_time
    api = DuplaApiBase(
        str(uuid.uuid4()),
        str(uuid.uuid4()),
        "pkcs12_filename",
        "pkcs12_password",
        "http://billetautomat.dk/url",
        5,
    )
    authentication_spy = mocker.spy(api, "_authenticate")

    api.get("http://some_api.dk/url")
    api.get("http://some_api.dk/url")

    authentication_spy.assert_called_once()
    mock_session_post.assert_called_once()
    assert mock_session_request.call_count == 2


def test_version_import():
    assert hasattr(dupla, "__version__")
    assert isinstance(dupla.__version__, str)
    assert dupla.__version__ == dupla.version.__version__
    assert dupla.__version__ == str(dupla.version.version_obj)



import random
import uuid
from datetime import date, datetime, timedelta
from typing import Type

import dateutil.parser
import pytest
from faker import Faker
from pydantic import BaseModel, ValidationError

import dupla as dp
from dupla.api_keys import DuplaApiKeys
from dupla.timestamp import TZ_UTC, as_utc

ALL_PAYLOADS: tuple[Type[dp.payload.BasePayload]] = (
    dp.payload.KtrPayload,
    dp.payload.LigPayload,
    dp.payload.MomsPayload,
    dp.payload.KtrObsPayload,
    dp.payload.LonsumPayload,
    dp.payload.SelskabSambeskatningPayload,
    dp.payload.SelskabSelvangivelsePayload,
    dp.payload.VirksomhedspligterPayload,
    dp.payload.VirksomhedsstatusPayload,
)

fake = Faker()


def cmp_datetime(dt1: datetime, dt2: datetime) -> None:
    assert isinstance(dt1, datetime)
    assert isinstance(dt2, datetime)
    dt1 = as_utc(dt1)
    dt2 = as_utc(dt2)
    assert dt1.date() == dt2.date()
    assert dt1.hour == dt2.hour
    assert dt1.minute == dt2.minute
    assert dt1.second == dt2.second


def build_dummy_api(max_tries=1, **kwargs) -> dp.DuplaAccess:
    api = dp.DuplaAccess(
        str(uuid.uuid4()),
        str(uuid.uuid4()),
        "pkcs12_filename",
        "pkcs12_password",
        r"http://billetautomat.dk/url",
        max_tries=max_tries,
        jwt_token_expiration_overlap=5,
        base_url=r"https://dummy.com",
        **kwargs,
    )
    return api


def get_num_of_len(n: int = 1):
    return fake.numerify("#" * n)


def get_fake_cvr(n=1):
    return [get_num_of_len(8) for _ in range(n)]


def get_fake_se(n=1):
    return get_fake_cvr(n=n)


def get_fake_cpr(n=1):
    return [get_num_of_len(10) for _ in range(n)]


def build_dummy_payload(api: Type[BaseModel]):
    k = dp.DuplaApiKeys  # Alias

    to_datetime = fake.date_time()
    from_datetime = to_datetime - timedelta(days=random.randint(1, 2 * 365))

    to_date = fake.date_object()
    from_date = to_date - timedelta(days=random.randint(1, 2 * 365))

    # Dummy datas
    data = {
        k.CVR: get_fake_cvr(n=random.randint(1, 15)),
        k.SE: get_fake_se(n=random.randint(1, 15)),
        k.CPR: get_fake_cpr(n=random.randint(1, 15)),
        k.AFREGNING_START: from_date,
        k.AFREGNING_SLUT: to_date,
        k.UDSTILLING_FRA: from_datetime,
        k.UDSTILLING_TIL: to_datetime,
        k.REGISTRERING_FORHOLD_FRA: from_date,
        k.REGISTRERING_FORHOLD_TIL: to_date,
        k.STATUS_GYLDIG_FRA: from_date,
        k.STATUS_GYLDIG_TIL: to_date,
        k.TEKNISK_REGISTRERING_FRA: from_date,
        k.TEKNISK_REGISTRERING_TIL: to_date,
        k.SELVANGIVESE_INDKOMST_AAR: str(random.randint(1970, 2050)),
    }

    def key_val_iter(api):
        for k, v in api.model_fields.items():
            key = k
            if v.alias:
                key = v.alias
            if key not in data:
                continue
            val = data[key]
            yield key, val

    payload = {key: val for key, val in key_val_iter(api) if key in data}
    return api(**payload)


@pytest.mark.parametrize("api", ALL_PAYLOADS)
def test_dummy_payload(api):
    m = build_dummy_payload(api)
    # Try building the payload
    payload = m.get_payload()
    assert "default_endpoint" not in payload


def test_get_payload(faker):
    to_date = faker.date_object()
    obj = dp.payload.KtrPayload(se=get_fake_se(), registrering_fra=to_date)
    payload = obj.get_payload()
    assert isinstance(payload, dict)
    assert all(isinstance(k, str) for k in payload.keys())
    assert DuplaApiKeys.SE in payload
    assert DuplaApiKeys.TEKNISK_REGISTRERING_FRA in payload
    # Unspecified, should not be in the payload
    assert DuplaApiKeys.CVR not in payload
    assert DuplaApiKeys.TEKNISK_REGISTRERING_TIL not in payload
    assert "default_endpoint" not in payload

    val = payload[dp.DuplaApiKeys.TEKNISK_REGISTRERING_FRA]
    assert isinstance(val, str)
    assert val == str(to_date)
    assert dateutil.parser.parse(val).date() == to_date


def test_payload_bad_arg(faker):
    with pytest.raises(ValidationError):
        dp.payload.KtrPayload(se=get_fake_se(), hello="123")


@pytest.mark.parametrize(
    "base_url",
    [
        r"https://api.skat.dk",
        r"https://api.tfe.skat.dk",
    ],
)
def test_moms_endpoint_url_build(base_url):
    cls = dp.payload.MomsPayload
    result = cls.endpoint_from_base_url(base_url)
    assert cls.default_endpoint == "Momsangivelse"
    assert r"/Momsangivelse" in result
    assert result == base_url + r"/Momsangivelse"


@pytest.mark.parametrize("payload", ALL_PAYLOADS)
@pytest.mark.parametrize(
    "base_url",
    [
        r"https://api.skat.dk",
        r"https://api.tfe.skat.dk",
    ],
)
def test_endpoint_url_build(base_url, payload: Type[dp.payload.BasePayload]):
    result = payload.endpoint_from_base_url(base_url)
    assert payload.default_endpoint in result
    assert result == base_url + r"/" + payload.default_endpoint


@pytest.mark.parametrize("as_iso", [False, True])
def test_moms_datetime(faker, as_iso: bool):
    """Test some date/datetime conversion"""
    for _ in range(20):
        # Repeat the test a few times with different dates
        dt: datetime = faker.date_time(tzinfo=TZ_UTC)
        obj = dp.payload.MomsPayload(
            se=get_fake_se(),
            udstilling_til=dt.isoformat() if as_iso else dt,
            afregning_start=dt.date(),
            afregning_slut=dt.date(),
        )

        assert obj.udstilling_til == dt
        payload = obj.get_payload()
        assert DuplaApiKeys.UDSTILLING_TIL in payload
        cmp_datetime(dateutil.parser.parse(payload[DuplaApiKeys.UDSTILLING_TIL]), dt)

        assert isinstance(payload[DuplaApiKeys.AFREGNING_START], str)
        assert payload[DuplaApiKeys.AFREGNING_START] == str(dt.date())


@pytest.mark.parametrize("payload_cls", ALL_PAYLOADS)
def test_api_builds_payload(payload_cls, mocker):
    spy = mocker.spy(payload_cls, "get_payload")
    spy2 = mocker.spy(payload_cls, "endpoint_from_base_url")
    api = build_dummy_api()
    obj = build_dummy_payload(payload_cls)
    assert spy.call_count == 0
    assert spy2.call_count == 0
    api.get_data(obj)
    assert spy.call_count == 1
    assert spy2.call_count == 1


@pytest.mark.parametrize("str_len", [0, 1, 2, 7, 9, 10])
def test_invalid_se(str_len: int):
    with pytest.raises(ValidationError):
        dp.payload.KtrPayload(se=[get_num_of_len(n=str_len)])
    dp.payload.KtrPayload(se=[get_num_of_len(n=8)])


def test_other_invalid_se():
    dp.payload.KtrPayload(se=[get_num_of_len(n=8)])
    with pytest.raises(ValidationError):
        dp.payload.KtrPayload(se=["a" * 8])


@pytest.mark.parametrize(
    "input",
    [
        ["12345678"],
        [12345678, 87654321],
        ["12345678", 12345678],
    ],
)
def test_int_or_string(input):
    dp.payload.KtrPayload(se=input)


def test_mom_udstilling(faker):
    payload = {
        DuplaApiKeys.SE: get_fake_se(n=5),
        DuplaApiKeys.AFREGNING_START: "1970-01-01",
        DuplaApiKeys.AFREGNING_SLUT: date.today(),
        DuplaApiKeys.UDSTILLING_FRA: faker.date_time(tzinfo=TZ_UTC),
        DuplaApiKeys.UDSTILLING_TIL: faker.date_time(tzinfo=TZ_UTC),
    }
    m = dp.payload.MomsPayload(**payload)
    assert isinstance(m, dp.payload.MomsPayload)
    assert isinstance(m.udstilling_til, datetime)
    assert isinstance(m.udstilling_fra, datetime)

    dct = m.get_payload()
    for key in [DuplaApiKeys.UDSTILLING_FRA, DuplaApiKeys.UDSTILLING_TIL]:
        assert isinstance(dct[key], str)
        fmt = dateutil.parser.parse(dct[key])
        assert fmt.tzname() == "UTC"
        exp = payload[key]
        cmp_datetime(fmt, exp)



from typing import Any

__all__ = ["DuplaApiException", "DuplaApiAuthenticationException", "InvalidPayloadException"]


class DuplaApiException(Exception):
    """A general error for the DUPLA API."""


class DuplaResponseException(DuplaApiException):
    """The Dupla response failed"""

    def __init__(self, *args, response: Any = None) -> None:
        self.response = response
        super().__init__(*args)


class DuplaApiAuthenticationException(Exception):
    """The authentication to the DUPLA API errored."""


class InvalidPayloadException(Exception):
    """Current payload is invalid for the payload schema."""



from pathlib import Path

from packaging.version import parse

__all__ = ["__version__"]

with Path(__file__).with_name("_version.txt").open("r") as file:
    version_obj = parse(file.readline().strip())

# string representation of the version
__version__ = str(version_obj)



from .version import *
from .base import *
from .endpoint import *
from .exceptions import *
from .api_keys import *

from . import payload

extra = ["payload"]

__all__ = version.__all__ + endpoint.__all__ + exceptions.__all__ + api_keys.__all__ + extra



from typing import Annotated, Any, Callable, List

from pydantic import AfterValidator, BeforeValidator


def _string_num_len_n(n: int) -> Callable[[str], str]:
    """Helper function to validate the correctness
    of integer-like strings. The length of the string is checked."""

    def _inner(s: str) -> str:
        try:
            int(s)  # Verify the number is int-like
        except ValueError:
            raise ValueError(f"Number must be integer-like, got {s!r}") from None
        if len(s) != n:
            raise ValueError(f"Must of length {n}.")
        return s

    return _inner


def _stringify_int(val: Any) -> Any:
    """Integers are valid, they just need to be strings"""
    if isinstance(val, int):
        return str(val)
    return val


CVR_STR = Annotated[str, BeforeValidator(_stringify_int), AfterValidator(_string_num_len_n(8))]
SE_STR = Annotated[str, BeforeValidator(_stringify_int), AfterValidator(_string_num_len_n(8))]
CPR_STR = Annotated[str, BeforeValidator(_stringify_int), AfterValidator(_string_num_len_n(10))]

CVR_T = List[CVR_STR]
CPR_T = List[CPR_STR]
SE_T = List[SE_STR]



from datetime import date, datetime
from typing import Union

ISO_FMT = "%Y-%m-%d"  # Date format for fields requiring yyyy-mm-dd


def convert_and_validate_iso_date(value: Union[str, date, datetime]) -> str:
    """Convert a date/datetime object to ISO.
    The value is validated to be a valid yyyy-mm-dd date string."""
    if isinstance(value, (date, datetime)):
        value = value.strftime(ISO_FMT)
    if not validate_iso_date(value):
        raise ValueError(f"Date '{value}' is not a date of the format yyyy-mm-dd")
    return value


def validate_iso_date(date: str) -> bool:
    """Verify that a date string has a valid correct yyyy-mm-dd format"""
    try:
        datetime.strptime(date, ISO_FMT)
    except ValueError:
        return False
    return True



from datetime import date, datetime

import pytest
from pydantic import ValidationError

from dupla.abstract_payload import BasePayload
from dupla.api_keys import DuplaApiKeys
from dupla.custom_types import CPR_T, CVR_T, SE_T
from dupla.payload import ENDP_T


class DummyBase(BasePayload):
    default_endpoint: ENDP_T = "dummy"


class DummySE(DummyBase):
    se: SE_T


class DummyCVR(DummyBase):
    cvr: CVR_T


class DummyCPR(DummyBase):
    cpr: CPR_T


class DummyDate(DummyBase):
    dt: date


class DummyDatetime(DummyBase):
    dt: datetime


def test_dummy_base():
    DummyBase()


@pytest.mark.parametrize(
    "input",
    [
        ["12345678"],
        [12345678, 87654321],
        ["12345678", 12345678],
    ],
)
@pytest.mark.parametrize("key", ["se", DuplaApiKeys.SE])
def test_int_se_conversion(input, key: str):
    kwargs = {key: input}
    inst = DummySE(**kwargs)
    assert len(inst.se) == len(input)
    assert all(isinstance(val, str) for val in inst.se)


@pytest.mark.parametrize(
    "input",
    [
        ["12345678"],
        [12345678, 87654321],
        ["12345678", 12345678],
    ],
)
@pytest.mark.parametrize("key", ["cvr", DuplaApiKeys.CVR])
def test_int_cvr_conversion(input, key):
    kwargs = {key: input}
    inst = DummyCVR(**kwargs)
    assert len(inst.cvr) == len(input)
    assert all(isinstance(val, str) for val in inst.cvr)


@pytest.mark.parametrize(
    "input",
    [
        ["1234567890"],
        [3103923289, 9876543210],
        ["1234567890", 1234567890],
    ],
)
@pytest.mark.parametrize("key", ["cpr", DuplaApiKeys.CPR])
def test_int_cpr_conversion(input, key):
    kwargs = {key: input}
    inst = DummyCPR(**kwargs)
    assert len(inst.cpr) == len(input)
    assert all(isinstance(val, str) for val in inst.cpr)


@pytest.mark.parametrize(
    "val",
    [
        "2023-10-01",
        date(year=2020, month=10, day=1),
    ],
)
def test_dummy_date(val):
    obj = DummyDate(dt=val)
    assert isinstance(obj.dt, date)


@pytest.mark.parametrize(
    "val",
    [
        "2023/10/03",
        "2023-10-01T20:00:00",
        "2023-10-01T20:00:00+00:00",
        "2023-10-01T20:00:00+01:00",
        "2023-10-01T20:00:00+02:00",
    ],
)
def test_invalid_date(val):
    with pytest.raises(ValidationError):
        DummyDate(dt=val)


@pytest.mark.parametrize(
    "val",
    [
        datetime.fromisoformat("2023-10-01T20:00:00+02:00"),
        "2023-10-01T20:00:00+02:00",
        "1900-01-01T20:00:00+00:00",
        date(year=2020, month=10, day=1),
    ],
)
def test_datetime(val):
    obj = DummyDatetime(dt=val)
    assert isinstance(obj.dt, datetime)


@pytest.mark.parametrize(
    "val",
    [
        "2023-10-01",
        "2023/10/03",
    ],
)
def test_invalid_datetime(val):
    with pytest.raises(ValidationError):
        DummyDatetime(dt=val)



import logging
from datetime import datetime, timedelta
from typing import Any, Dict, Optional
from uuid import uuid4

import requests
import requests_pkcs12

from .exceptions import DuplaApiAuthenticationException
from .timestamp import get_utc_now

__all__ = [
    "DuplaApiBase",
]

logger = logging.getLogger(__file__)


class DuplaApiBase:
    """Base class for API acces to the Dataudveklspingsplatformen API (Dupla).
    Handles authentication and headers for the API requests.

    Arguments:
        transaction_id (str): An ID used to correlate requests across the API.
            Should be constant for IKP-DA.
        agreement_id (str): An ID/token supplied by the API provider.
        pkcs12_filename (str): Path to PKCS12 certificate file.
        pkcs12_password (str): Password for PKCS12 certificate file.
        billetautomat_url (str): Endpoint to the authentication service for requesting JWT tokens.
        jwt_token_expiration_overlap (int): The overlap time for token expiration time (in seconds)
            to avoid situations where token is almost expired during the check and will be rejected
            in a next request.
    """

    transaction_id: str
    agreement_id: str
    _pkcs12_adapter: requests_pkcs12.Pkcs12Adapter

    def __init__(
        self,
        transaction_id: str,
        agreement_id: str,
        pkcs12_filename: str,
        pkcs12_password: str,
        billetautomat_url: str,
        jwt_token_expiration_overlap: int,
    ):
        self.transaction_id = transaction_id
        self.agreement_id = agreement_id

        self._pkcs12_adapter = requests_pkcs12.Pkcs12Adapter(
            pkcs12_filename=pkcs12_filename,
            pkcs12_password=pkcs12_password,
        )
        self.billetautomat_url = billetautomat_url
        self.jwt_token_expiration_overlap = jwt_token_expiration_overlap
        self.token_expiration_time: Optional[datetime] = None
        self.jwt_token: Optional[str] = None

    def request(self, method, url, **kwargs) -> requests.Response:
        """Constructs and sends a `requests.Request` with appropriate headers
        (including the JWT authenticationtoken) for the Dupla API.
        If token is not present or is expired - first sends a request to the authentication
        service using mTlS connection and the set certificate to get the JWT authentication token.
        Then the token is used to authenticate requests to Dupla API until it's expired -
        then the JWT token is retrieved again.

        Arguments:
            method (str): HTTP method of the `requests.Request` object
            url (str): URL for the new :class:`Request` object.
            **kwargs (Optional[Dict]): Optional arguments that `requests.request` takes.

        Returns:
            requests.Reponse: A requests Response opject
        """
        request_id = uuid4()
        if not self._is_token_present() or self._is_token_expired():
            self._authenticate()

        headers = {
            "X-Request-ID": str(request_id),
            "X-Transaktions-ID": self.transaction_id,
            "UFST-Adgangsgrundlag": f"urn:ufst:adgangsgrundlag:aftale:{self.agreement_id}",
            "Authorization": f"Bearer {self.jwt_token}",
        }

        with requests.Session() as session:
            session.headers.update(headers)

            return session.request(method, url, **kwargs)

    def get(
        self, url: str, params: Optional[Dict] = None, **kwargs: Dict[str, Any]
    ) -> requests.Response:
        """Sends a GET request, handling authentication and API headers.

        Arguments:
            url (str): URL for the new `requests.Request` object.
            params (Optional[Dict]): Dictionary, list of tuples or bytes to send in the
                query string for the :class:`Request`.
            **kwargs (Optional[Dict]): Optional arguments that `requests.request` takes.

        Returns:
            requests.Reponse: A requests Response opject
        """
        return self.request("get", url, params=params, **kwargs)

    def _authenticate(self) -> None:
        """Retrieves a JWT token from the authentication service (billetautomat) to be used for
        Dupla API requests. The connection to the authentication service is encrypted with mTLS
        and the set certificate. To retrieve a JWT token, the payload must include 2 fields -
        client_id which is always "api-gateway" and so called "nonce" which is uuid v4 string.
        Nonce should be unique per request and can be understood as a request id.
        The token expiration time is set as well for further checks.
        """
        self.token_expiration_time = None
        self.jwt_token = None
        nonce = str(uuid4())

        payload = {"client_id": "api-gateway", "nonce": nonce}

        with requests.Session() as session:
            session.mount(self.billetautomat_url, self._pkcs12_adapter)
            result = session.post(self.billetautomat_url, json=payload)
            if result.ok:
                result_payload = result.json()
            else:
                raise DuplaApiAuthenticationException(
                    f"JWT error: fetching the token failed, "
                    f"http code: {result.status_code}, "
                    f"message: {result.content.decode()}"
                )

        if access_token := result_payload.get("access_token"):
            self.jwt_token = access_token
        else:
            raise DuplaApiAuthenticationException(
                "JWT error: access_token not present in response payload"
            )
        if expires_in := result_payload.get("expires_in"):
            self.token_expiration_time = get_utc_now() + timedelta(seconds=expires_in)
        else:
            raise DuplaApiAuthenticationException(
                "JWT error: expires_in not present in response payload"
            )

    def _is_token_present(self) -> bool:
        """Checks whether the JWT token was retrieved and the expiration time is set.

        Returns:
            True if the JWT token and expiration time are set, False otherwise.
        """
        return self.jwt_token is not None and self.token_expiration_time is not None

    def _is_token_expired(self) -> bool:
        """Checks whether the JWT token is expired. To avoid situations where the token is valid
        for the last second but could be considered as expired in a next request - overlap time
        is used to consider the token as expired quicker.

        Returns:
            True if the JWT token is expired, False otherwise.
        """
        return get_utc_now() >= self.token_expiration_time - timedelta(
            seconds=self.jwt_token_expiration_overlap
        )



import zoneinfo
from datetime import datetime

TZ_UTC = zoneinfo.ZoneInfo("UTC")
TZ_DK = zoneinfo.ZoneInfo("Europe/Copenhagen")


def get_utc_now() -> datetime:
    return datetime.now(tz=TZ_UTC)


def get_dk_now() -> datetime:
    return datetime.now(tz=TZ_DK)


def as_utc(ts: datetime) -> datetime:
    return ts.astimezone(TZ_UTC)
